/*
fonthandlerimpl.cpp -- Font subsetize tool for SDL_Pango.
Copyright (C) 2004 NAKAMURA Ken'ichi

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <iostream>
#include <fstream>
#include <map>
#include <set>
#include <deque>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include <xercesc/framework/XMLFormatter.hpp>
#include <boost/python.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/function.hpp>
#include <boost/format.hpp>

using xercesc::XMLPlatformUtils;
using xercesc::XMLException;
using xercesc::SAXParser;
using xercesc::SAXParseException;
using xercesc::XMLString;
using xercesc::HandlerBase;
using xercesc::AttributeList;
using xercesc::XMLFormatTarget;
using xercesc::XMLFormatter;
using boost::python::object;
using boost::python::class_;
using boost::shared_ptr;
using boost::function;
using boost::python::extract;
using std::string;
typedef std::ostream::pos_type pos_type;

inline string transcode(const XMLCh* const xmlch) {
    string result;
    if ( xmlch ) {
	char* p = XMLString::transcode(xmlch);
	result = p;
	XMLString::release(&p);
    }
    return result;
}

class HandlerBaseEx : public HandlerBase
{
public:
    typedef shared_ptr<HandlerBaseEx> HandlerBaseEx_sp;
    HandlerBaseEx_sp m_ParentHandler;
    void setParentHandler(HandlerBaseEx_sp parentHandler)
    {
	m_ParentHandler = parentHandler;
    }
};

typedef std::deque<string> ElementPath;

typedef shared_ptr<HandlerBaseEx> HandlerBaseEx_sp;
class HandlerTreeNode : public std::map<string, HandlerTreeNode>
{
    HandlerBaseEx_sp m_Handler;

public:
    HandlerTreeNode(){}

    HandlerTreeNode(HandlerBaseEx_sp handler)
    {
	m_Handler = handler;
    }

    HandlerBaseEx_sp GetHandler(const ElementPath& elementPath, int level = 0) const
    {
	if (elementPath.size() == level)
	    return m_Handler;

	HandlerTreeNode::const_iterator i = this->find(elementPath[level]);
	if (i == this->end())
	    return m_Handler;

	i->second.m_Handler->setParentHandler(m_Handler);
	return i->second.GetHandler(elementPath, level + 1);
    }
};

namespace XERCES_CPP_NAMESPACE {
    static const XMLCh  gEndElement[] = { chOpenAngle, chForwardSlash, chNull };
    static const XMLCh  gEndPI[] = { chQuestion, chCloseAngle, chNull };
    static const XMLCh  gStartPI[] = { chOpenAngle, chQuestion, chNull };
    static const XMLCh  gXMLDecl1[] = {
	chOpenAngle, chQuestion, chLatin_x, chLatin_m, chLatin_l,
	    chSpace, chLatin_v, chLatin_e, chLatin_r, chLatin_s, chLatin_i,
	    chLatin_o, chLatin_n, chEqual, chDoubleQuote, chDigit_1, chPeriod,
	    chDigit_0, chDoubleQuote, chSpace, chLatin_e, chLatin_n, chLatin_c,
	    chLatin_o, chLatin_d, chLatin_i, chLatin_n, chLatin_g, chEqual,
	    chDoubleQuote, chNull
    };

    static const XMLCh  gXMLDecl2[] = {
	chDoubleQuote, chQuestion, chCloseAngle, chLF, chNull
    };
}
using xercesc::gXMLDecl1;
using xercesc::gXMLDecl2;
using xercesc::gEndElement;
using xercesc::gStartPI;
using xercesc::gEndPI;

typedef shared_ptr<HandlerTreeNode> HandlerTreeNode_sp;
class SAXHandlers;
HandlerTreeNode_sp CreateHandlerTreeRoot(SAXHandlers* inst);

class SAXHandlers : public HandlerBaseEx
{
private :
    // -----------------------------------------------------------------------
    //      This is the formatter object that is used to output the data
    //      to the target. It is set up to format to the standard output
    //      stream.
    // -----------------------------------------------------------------------
    class TTXFormatTarget : public XMLFormatTarget
    {
    public:
	std::ofstream m_fout;
	TTXFormatTarget(string filename)
	{
	    m_fout.open(filename.c_str());
	}

	~TTXFormatTarget()
	{
	    m_fout.flush();
	    m_fout.close();
	}

	// ---------------------------------------------------------------------------
	//  Overrides of the output formatter target interface
	// ---------------------------------------------------------------------------
	void writeChars(const XMLByte* const toWrite)
	{
	}

	void writeChars(const XMLByte* const toWrite,
	    const unsigned int count,
	    XMLFormatter* const formatter)
	{
	    // For this one, just dump them to the standard output
	    // Surprisingly, Solaris was the only platform on which
	    // required the char* cast to print out the string correctly.
	    // Without the cast, it was printing the pointer value in hex.
	    // Quite annoying, considering every other platform printed
	    // the string with the explicit cast to char* below.
	    m_fout.write((char *) toWrite, (int) count);
	}
    };
    typedef shared_ptr<TTXFormatTarget> TTXFormatTarget_sp;
    TTXFormatTarget_sp m_FormatTarget;

    ElementPath m_ElementPath;
    HandlerTreeNode_sp m_HandlerRootNode;

    HandlerBaseEx_sp GetHandler()
    {
	return m_HandlerRootNode->GetHandler(m_ElementPath);
    }

public:
    SAXHandlers()
    {
    }

    // ---------------------------------------
    // implements
    // ---------------------------------------
public:
    typedef shared_ptr<XMLFormatter> XMLFormatter_sp;
    XMLFormatter_sp m_Formatter;

    typedef std::map<int, string> GlyphIDtoNameMap;
    GlyphIDtoNameMap m_GlyphIDtoNameMap;

    typedef std::set<string> UsedGlyphName;
    UsedGlyphName m_UsedGlyphName;

    bool isUsedGlyphName(string name)
    {
	UsedGlyphName::iterator i = m_UsedGlyphName.find(name);
	if (i == m_UsedGlyphName.end())
	    return false;
	return true;
    }

    object m_Codebook;
    void setCodebook(object codebook)
    {
	m_Codebook = codebook;
    }

    object m_Chapter;
    void setChapter(object chapter)
    {
	m_Chapter = chapter;
    }

    string m_OutputFilename;
    void setOutputFilename(string filename)
    {
	m_OutputFilename = filename;
    }

    pos_type tellp()
    {
	return m_FormatTarget->m_fout.tellp();
    }

    void seekp(pos_type p)
    {
	m_FormatTarget->m_fout.seekp(p);
    }

public:
    // ---------------------------------------
    // content handler
    // ---------------------------------------
    void startDocument()
    {
	m_FormatTarget = TTXFormatTarget_sp(new TTXFormatTarget(m_OutputFilename));

	m_Formatter = XMLFormatter_sp(new XMLFormatter("ISO-8859-1", 0, &(*m_FormatTarget),
	    XMLFormatter::NoEscapes, XMLFormatter::UnRep_CharRef));

	m_HandlerRootNode = CreateHandlerTreeRoot(this);

	*m_Formatter << gXMLDecl1 << m_Formatter->getEncodingName() << gXMLDecl2;
    }

    void startElement(const XMLCh* const name, AttributeList&  attributes)
    {
	m_ElementPath.push_back(transcode(name));

	GetHandler()->startElement(name, attributes);
    }

    void endElement(const XMLCh* const name)
    {
	GetHandler()->endElement(name);

	m_ElementPath.pop_back();
    }

    void characters(const XMLCh* const chars, const unsigned int length)
    {
	GetHandler()->characters(chars, length);
    }

    void ignorableWhitespace(const XMLCh* const chars, const unsigned int length)
    {
	GetHandler()->ignorableWhitespace(chars, length);
    }

    void processingInstruction(const  XMLCh* const target, const XMLCh* const data)
    {
	GetHandler()->processingInstruction(target, data);
    }

    void resetDocument()
    {
    }

    // ---------------------------------------
    // error handler
    // ---------------------------------------
    void error(const SAXParseException& e)
    {
    }

    void fatalError(const SAXParseException& e)
    {
    }

    void warning(const SAXParseException& e)
    {
    }

    void resetErrors()
    {
    }
};

class SAXHandlerDelegateBase : public HandlerBaseEx
{
protected:
    SAXHandlers* m_Base;

public:
    SAXHandlerDelegateBase(SAXHandlers* base)
    {
	m_Base = base;
    }
};

class CopyHandler : public SAXHandlerDelegateBase
{
public:
    CopyHandler(SAXHandlers* base) : SAXHandlerDelegateBase(base){}

    void startElement(const XMLCh* const name, AttributeList& attributes)
    {
	// The name has to be representable without any escapes
	*m_Base->m_Formatter  << XMLFormatter::NoEscapes << xercesc::chOpenAngle << name;

	unsigned int len = attributes.getLength();
	for (unsigned int index = 0; index < len; index++) {

	    //  Again the name has to be completely representable. But the
	    //  attribute can have refs and requires the attribute style
	    //  escaping.

	    *m_Base->m_Formatter  << XMLFormatter::NoEscapes
		<< xercesc::chSpace << attributes.getName(index)
		<< xercesc::chEqual << xercesc::chDoubleQuote
		<< XMLFormatter::AttrEscapes
		<< attributes.getValue(index)
		<< XMLFormatter::NoEscapes
		<< xercesc::chDoubleQuote;
	}
	*m_Base->m_Formatter << xercesc::chCloseAngle;
    }

    void endElement(const XMLCh* const name)
    {
	// No escapes are legal here
	*m_Base->m_Formatter << XMLFormatter::NoEscapes << gEndElement << name << xercesc::chCloseAngle;
    }

    void characters(const XMLCh* const chars, const unsigned int length)
    {
	m_Base->m_Formatter->formatBuf(chars, length, XMLFormatter::CharEscapes);
    }

    void ignorableWhitespace(const XMLCh* const chars, const unsigned int length)
    {
	m_Base->m_Formatter->formatBuf(chars, length, XMLFormatter::NoEscapes);
    }

    void processingInstruction(const  XMLCh* const target, const XMLCh* const data)
    {
	*m_Base->m_Formatter << XMLFormatter::NoEscapes << gStartPI  << target;
	if (data)
	    *m_Base->m_Formatter << xercesc::chSpace << data;
	*m_Base->m_Formatter << XMLFormatter::NoEscapes << gEndPI;
    }
};

class DeleteHandler : public SAXHandlerDelegateBase
{
public:
    DeleteHandler(SAXHandlers* base) : SAXHandlerDelegateBase(base){}

    void startElement(const XMLCh* const name, AttributeList&  attributes)
    {
    }

    void endElement(const XMLCh* const name)
    {
    }

    void characters(const XMLCh* const chars, const unsigned int length)
    {
    }

    void ignorableWhitespace(const XMLCh* const chars, const unsigned int length)
    {
    }
};

// ---------------------------------------
// handler delegate base class with utility
// ---------------------------------------
class SAXHandlerDelegateBaseEx : public SAXHandlerDelegateBase
{
protected:
    HandlerBaseEx_sp m_CopyHandler;
    bool m_IsLastElementCopied;

    long parseHexOfAttr(const char* attrName, AttributeList&  attributes)
    {
	const XMLCh* content_str = attributes.getValue(attrName);
	assert(content_str);
	char *p = XMLString::transcode(content_str);
	long v = strtol(p, NULL, 16);
	XMLString::release(&p);

	return v;
    }

    void outputAttribute(string name, string value)
    {
	XMLCh* value_x = XMLString::transcode(value.c_str());
	outputAttribute(name, value_x);
	XMLString::release(&value_x);
    }

    void outputAttribute(string name, const XMLCh* value)
    {
	XMLCh* name_x = XMLString::transcode(name.c_str());
	outputAttribute(name_x, value);
	XMLString::release(&name_x);
    }

    void outputAttribute(const XMLCh* name, const XMLCh* value)
    {
	*m_Base->m_Formatter  << XMLFormatter::NoEscapes
	    << xercesc::chSpace << name
	    << xercesc::chEqual << xercesc::chDoubleQuote
	    << XMLFormatter::AttrEscapes
	    << value << XMLFormatter::NoEscapes
	    << xercesc::chDoubleQuote;
    }

    void beforeOutputAttribute(const XMLCh* name)
    {
	// The name has to be representable without any escapes
	*m_Base->m_Formatter  << XMLFormatter::NoEscapes << xercesc::chOpenAngle << name;
    }

    void afterOutputAttribute()
    {
	*m_Base->m_Formatter << xercesc::chCloseAngle;
    }

public:
    SAXHandlerDelegateBaseEx(SAXHandlers* base) : SAXHandlerDelegateBase(base)
    {
	m_CopyHandler = HandlerBaseEx_sp(new CopyHandler(base));
	m_IsLastElementCopied = false;
    }

    bool startElementCopyIfUsedGlyphName(string glyphName, const XMLCh* const name, AttributeList&  attributes)
    {
	if (m_Base->isUsedGlyphName(glyphName)) {
	    m_CopyHandler->startElement(name, attributes);
	    m_IsLastElementCopied = true;
	} else {
	    m_IsLastElementCopied = false;
	}
	return m_IsLastElementCopied;
    }

    void endElement(const XMLCh* const name)
    {
	if (m_IsLastElementCopied)
	    m_CopyHandler->endElement(name);
    }

    void characters(const XMLCh* const chars, const unsigned int length)
    {
	if (m_IsLastElementCopied)
	    m_CopyHandler->characters(chars, length);
    }

    void ignorableWhitespace(const XMLCh* const chars, const unsigned int length)
    {
	if (m_IsLastElementCopied)
	    m_CopyHandler->ignorableWhitespace(chars, length);
    }
};

// ---------------------------------------
// ttFont/GlyphOrder/GlyphID handler
// ---------------------------------------
class GlyphIDHandler : public SAXHandlerDelegateBaseEx
{
    function<bool (int glyphCode)> m_HasGlyph;

public:
    GlyphIDHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base)
    {
	m_HasGlyph = base->m_Chapter.attr("hasGlyph");
    }

    void startElement(const XMLCh* const name, AttributeList&  attributes)
    {
	const XMLCh* id_content_str = attributes.getValue("id");
	assert(id_content_str);
	int id_content = XMLString::parseInt(id_content_str);

	if (id_content == 0) {
	    // 0 is .notdef
	    m_Base->m_Chapter.attr("addUsedGlyphs")(boost::python::make_tuple(0));
	}

	const XMLCh* name_content_str = attributes.getValue("name");
	assert(name_content_str);
	string name_content = transcode(name_content_str);
	
	if (m_HasGlyph(id_content)) {
	    m_Base->m_GlyphIDtoNameMap[id_content] = name_content;
	    m_Base->m_UsedGlyphName.insert(name_content);
	    m_CopyHandler->startElement(name, attributes);
	    m_IsLastElementCopied = true;
	} else {
	    m_IsLastElementCopied = false;
	}
    }
};

// ---------------------------------------
// ttFont/OS_2/fsType handler
// ---------------------------------------
class FsTypeHandler : public SAXHandlerDelegateBaseEx
{
public:
    FsTypeHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base){}

    void startElement(const XMLCh* const name, AttributeList&  attributes)
    {
	const XMLCh* value_content_str = attributes.getValue("value");
	string value_content = transcode(value_content_str);
	string value_hex = value_content.substr(9, 8);
	long v = strtol(value_hex.c_str(), NULL, 16);

	if (v == 2)
	    throw 1;	// Restricted License embedding
	else if (v & 0x200)
	    throw 1;	// Bitmap embedding only
	m_CopyHandler->startElement(name, attributes);
	m_IsLastElementCopied = true;
    }
};

// ---------------------------------------
// ttFont/GSUB/LookupList/Lookup/SingleSubst handler
// ---------------------------------------
class SingleSubstHandler : public SAXHandlerDelegateBaseEx
{
    pos_type m_BeforeStartElement;
    bool m_HasSubstitutionElement;

public:
    SingleSubstHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base){}

    void startElement(const XMLCh* const name, AttributeList&  attributes)
    {
	m_BeforeStartElement = m_Base->tellp();
	m_CopyHandler->startElement(name, attributes);
	m_HasSubstitutionElement = false;
    }

    void hasSubstitutionElement()
    {
	m_HasSubstitutionElement = true;
    }

    void endElement(const XMLCh* const name)
    {
	if (m_HasSubstitutionElement) {
	    m_CopyHandler->endElement(name);
	} else {
	    m_Base->seekp(m_BeforeStartElement);
	}
    }
};

// ---------------------------------------
// ttFont/GSUB/LookupList/Lookup/SingleSubst/Substitution handler
// ---------------------------------------
class SingleSubstitutionHandler : public SAXHandlerDelegateBaseEx
{
public:
    SingleSubstitutionHandler(SAXHandlers* base): SAXHandlerDelegateBaseEx(base) {}

    void startElement(const XMLCh* const name, AttributeList&  attributes)
    {
	const XMLCh* out_content_str = attributes.getValue("out");
	assert(out_content_str);
	string out_content = transcode(out_content_str);

	bool isCopied = startElementCopyIfUsedGlyphName(out_content, name, attributes);
	if (isCopied) {
	    SingleSubstHandler* ssh = dynamic_cast<SingleSubstHandler*>(&(*m_ParentHandler));
	    ssh->hasSubstitutionElement();
	}
    }
};

// ---------------------------------------
// ttFont/cmap/cmap_format_0/map handler
// ---------------------------------------
class CmapFormat0MapHandler : public SAXHandlerDelegateBaseEx
{
    string m_Notdef_str;

public:
    CmapFormat0MapHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base){}

    void startElement(const XMLCh* const name, AttributeList&  attributes)
    {
	static string name_label = "name";
	static string code_label = "code";

	long code_content = parseHexOfAttr("code", attributes);
	if (code_content == 0) {
	    m_Notdef_str = m_Base->m_GlyphIDtoNameMap[0];
	}

	beforeOutputAttribute(name);

	outputAttribute(code_label, attributes.getValue("code"));
	outputAttribute(name_label, m_Notdef_str);

	afterOutputAttribute();

	m_IsLastElementCopied = true;
    }
};

// ---------------------------------------
// ttFont/cmap/cmap_format_4/map handler
// ---------------------------------------
class CmapFormat4MapHandler : public SAXHandlerDelegateBaseEx
{
    object m_GetCipheredChar;
    int getCipheredChar(int glyphCode)
    {
	return extract<int>(m_GetCipheredChar(glyphCode));
    }

public:
    CmapFormat4MapHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base)
    {
	m_GetCipheredChar = m_Base->m_Codebook.attr("getCipheredChar");
    }

    void startElement(const XMLCh* const name, AttributeList&  attributes)
    {
	static string name_label = "name";
	static string code_label = "code";

	const XMLCh* name_content_str = attributes.getValue("name");
	assert(name_content_str);
	string name_content = transcode(name_content_str);

	if (m_Base->isUsedGlyphName(name_content)) {
	    long code_content = parseHexOfAttr("code", attributes);
	    int code_ciphered = getCipheredChar(code_content);
	    string code_ciphered_hex = boost::io::str(boost::format("0x%x") % code_ciphered);

	    beforeOutputAttribute(name);

	    outputAttribute(code_label, code_ciphered_hex);
	    outputAttribute(name_label, name_content_str);

	    afterOutputAttribute();

	    m_IsLastElementCopied = true;
	} else {
	    m_IsLastElementCopied = false;
	}
    }

};

// ---------------------------------------
// ttFont/glyf/TTGlyph handler
// ---------------------------------------
class TTGlyphHandler : public SAXHandlerDelegateBaseEx
{
public:
    TTGlyphHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base){}

    void startElement(const XMLCh* const name, AttributeList& attributes)
    {
	static string TTGlyph_label = "TTGlyph";
	string element = transcode(name);
	if (element == TTGlyph_label) {
	    string name_content = transcode(attributes.getValue("name"));
	    startElementCopyIfUsedGlyphName(name_content, name, attributes);
	} else if (m_IsLastElementCopied) {
	    m_CopyHandler->startElement(name, attributes);
	}
    }
};

// ---------------------------------------
// ttFont/hmtx/mtx handler
// ttFont/vmtx/mtx handler
// ---------------------------------------
class MtxHandler : public SAXHandlerDelegateBaseEx
{
public:
    MtxHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base) {}

    void startElement(const XMLCh* const name, AttributeList& attributes)
    {
	const XMLCh* name_content_str = attributes.getValue("name");
	assert(name_content_str);
	string name_content = transcode(name_content_str);
	startElementCopyIfUsedGlyphName(name_content, name, attributes);
    }
};

// ---------------------------------------
// ttFont/name/namerecord handler
// ---------------------------------------
class NameRecordHandler : public SAXHandlerDelegateBaseEx
{
    bool m_IsSubstitutionMode;
    object m_GetSubstituteFace;

    int parseIntOfAttr(const char* attrName, AttributeList& attributes)
    {
	const XMLCh* content_str = attributes.getValue(attrName);
	assert(content_str);
	return XMLString::parseInt(content_str);
    }

    string getSubstituteFace(string originalFace)
    {
	return extract<string>(m_GetSubstituteFace(originalFace.c_str()));
    }

public:
    NameRecordHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base)
    {
	m_IsSubstitutionMode = false;
	m_GetSubstituteFace = base->m_Codebook.attr("getSubstituteFace");
    }

    void startElement(const XMLCh* const name, AttributeList& attributes)
    {
	int nameID = parseIntOfAttr("nameID", attributes);
	int platformID = parseIntOfAttr("platformID", attributes);
	int platEncID = parseIntOfAttr("platEncID", attributes);
	long langID = parseHexOfAttr("langID", attributes);

	if (nameID == 1) {
	    if (langID == 0 || langID == 1033) {
		// Latin
		m_IsSubstitutionMode = true;
	    } else if (platformID == 3 && platEncID == 1) {
		// Unicode UTF16
		// TODO implement me
		m_IsSubstitutionMode = false;
	    } else if (platformID == 3 && platEncID == 2) {
		// ShiftJIS
		// TODO implement me
		m_IsSubstitutionMode = false;
	    } else {
		m_IsSubstitutionMode = false;
	    }
	} else {
	    m_IsSubstitutionMode = false;
	}

	m_CopyHandler->startElement(name, attributes);
	m_IsLastElementCopied = true;
    }

    void characters(const XMLCh* const chars, const unsigned int length)
    {
	if (! m_IsSubstitutionMode) {
	    m_CopyHandler->characters(chars, length);
	    return;
	}

	XMLCh* raw = XMLString::replicate(chars);
	XMLString::trim(raw);
	string originalFace = transcode(raw);
	XMLString::release(&raw);

	string substituteFace = getSubstituteFace(originalFace);
	XMLCh* newChars = XMLString::transcode(substituteFace.c_str());
	unsigned int newLength = XMLString::stringLen(newChars);
	m_CopyHandler->characters(newChars, newLength);
	XMLString::release(&newChars);
    }
};

// ---------------------------------------
// ttFont/LTSH/yPel handler
// ---------------------------------------
class LtshYpelHandler : public SAXHandlerDelegateBaseEx
{
public:
    LtshYpelHandler(SAXHandlers* base) : SAXHandlerDelegateBaseEx(base) {}

    void startElement(const XMLCh* const name, AttributeList& attributes)
    {
	const XMLCh* name_content_str = attributes.getValue("name");
	assert(name_content_str);
	string name_content = transcode(name_content_str);
	startElementCopyIfUsedGlyphName(name_content, name, attributes);
    }
};


HandlerTreeNode_sp CreateHandlerTreeRoot(SAXHandlers* inst)
{
    HandlerBaseEx_sp copyHandler = HandlerBaseEx_sp(new CopyHandler(inst));
    HandlerBaseEx_sp deleteHandler = HandlerBaseEx_sp(new DeleteHandler(inst));

    HandlerTreeNode copyNode(copyHandler);

    HandlerTreeNode deleteNode(deleteHandler);

    HandlerTreeNode_sp root = HandlerTreeNode_sp(new HandlerTreeNode(copyHandler));

    HandlerTreeNode glyphID(HandlerBaseEx_sp(new GlyphIDHandler(inst)));
    HandlerTreeNode glyphOrder(copyHandler);
    glyphOrder["GlyphID"] = glyphID;
    HandlerTreeNode ttFont(copyHandler);
    ttFont["GlyphOrder"] = glyphOrder;

    HandlerTreeNode fsType(HandlerBaseEx_sp(new FsTypeHandler(inst)));
    HandlerTreeNode os2(copyHandler);
    os2["fsType"] = fsType;
    ttFont["OS_2"] = os2;

    ttFont["EBDT"] = deleteNode;
    ttFont["EBLC"] = deleteNode;
    ttFont["EBSC"] = deleteNode;

    HandlerTreeNode singleSubstitution(HandlerBaseEx_sp(new SingleSubstitutionHandler(inst)));
    HandlerTreeNode singleSubst(HandlerBaseEx_sp(new SingleSubstHandler(inst)));
    singleSubst["Substitution"] = singleSubstitution;
    HandlerTreeNode lookup(copyHandler);
    lookup["SingleSubst"] = singleSubst;
    HandlerTreeNode lookupList(copyHandler);
    lookupList["Lookup"] = lookup;
    HandlerTreeNode gsub(copyHandler);
    gsub["LookupList"] = lookupList;
    ttFont["GSUB"] = gsub;

    HandlerTreeNode cmap_format_0_map(HandlerBaseEx_sp(new CmapFormat0MapHandler(inst)));
    HandlerTreeNode cmap_format_0(copyHandler);
    cmap_format_0["map"] = cmap_format_0_map;
    HandlerTreeNode cmap_format_4_map(HandlerBaseEx_sp(new CmapFormat4MapHandler(inst)));
    HandlerTreeNode cmap_format_4(copyHandler);
    cmap_format_4["map"] = cmap_format_4_map;
    HandlerTreeNode cmap(copyHandler);
    cmap["cmap_format_0"] = cmap_format_0;
    cmap["cmap_format_4"] = cmap_format_4;
    ttFont["cmap"] = cmap;

    HandlerTreeNode ttglyph(HandlerBaseEx_sp(new TTGlyphHandler(inst)));
    HandlerTreeNode glyf(copyHandler);
    glyf["TTGlyph"] = ttglyph;
    ttFont["glyf"] = glyf;

    HandlerTreeNode mtx(HandlerBaseEx_sp(new MtxHandler(inst)));
    HandlerTreeNode hmtx(copyHandler);
    hmtx["mtx"] = mtx;
    ttFont["hmtx"] = hmtx;

    ttFont["mort"] = deleteHandler;

    HandlerTreeNode ltsh_ypel(HandlerBaseEx_sp(new LtshYpelHandler(inst)));
    HandlerTreeNode ltsh(copyHandler);
    ltsh["yPel"] = ltsh_ypel;
    ttFont["LTSH"] = ltsh;

    HandlerTreeNode nameRecord(HandlerBaseEx_sp(new NameRecordHandler(inst)));
    HandlerTreeNode name(copyHandler);
    name["namerecord"] = nameRecord;
    ttFont["name"] = name;

    HandlerTreeNode vmtx(copyHandler);
    vmtx["mtx"] = mtx;
    ttFont["vmtx"] = vmtx;

    ttFont["post"] = deleteHandler;

    HandlerTreeNode visualnovel(deleteHandler);

    (*root)[string("ttFont")] = ttFont;
    (*root)[string("visualnovel")] = visualnovel;

    return root;
}


struct FontHandlerImpl
{
    int run(string in_filename, string out_filename, object codebook, object chapter)
    {
	SAXParser parser;
	SAXHandlers handler;
	handler.setCodebook(codebook);
	handler.setChapter(chapter);
	handler.setOutputFilename(out_filename);
	parser.setDocumentHandler(&handler);
	parser.setErrorHandler(&handler);

	try {
	    parser.parse(in_filename.c_str());
	} catch (const XMLException& toCatch) {
	    char* message = XMLString::transcode(toCatch.getMessage());
	    std::cout << "Exception message is: \n"
		<< message << "\n";
	    XMLString::release(&message);
	    return -1;
	} catch (const SAXParseException& toCatch) {
	    char* message = XMLString::transcode(toCatch.getMessage());
	    std::cout << "Exception message is: \n"
		<< message << "\n";
	    XMLString::release(&message);
	    return -1;
	} catch (int returnCode) {
	    return returnCode;
	}/* catch (...) {
	    std::cout << "Unexpected Exception \n" ;
	    return -1;
	}*/

	return 0;
    }
};

BOOST_PYTHON_MODULE(fonthandlerimpl)
{
    XMLPlatformUtils::Initialize();
    class_<FontHandlerImpl>("FontHandlerImpl")
		.def("run", &FontHandlerImpl::run)
    ;
}
